<?php

namespace Oro\Bundle\SalesBundle\Entity\Repository;

use Doctrine\DBAL\Types\Types;
use Doctrine\ORM\EntityRepository;
use Doctrine\ORM\Query;
use Doctrine\ORM\QueryBuilder;
use Oro\Bundle\SecurityBundle\ORM\Walker\AclHelper;
use Oro\Bundle\WorkflowBundle\Helper\WorkflowQueryTrait;
use Oro\Bundle\WorkflowBundle\Model\Workflow;
use Oro\Component\DoctrineUtils\ORM\QueryBuilderUtil;

/**
 * Doctrine repository for SalesFunnel entity.
 * @deprecated since 2.0 will be removed after 2.2
 */
class SalesFunnelRepository extends EntityRepository
{
    use WorkflowQueryTrait;
    /**
     * @var array
     */
    protected $excludedSteps = [
        'new_lead',
        'disqualified_lead',
        'lost_opportunity',
    ];

    /**
     * @param  \DateTime $dateFrom
     * @param  \DateTime $dateTo
     * @param  Workflow  $workflow
     * @param  array     $customStepCalculations
     * @param  AclHelper $aclHelper
     *
     * @return array
     * @SuppressWarnings(PHPMD.CyclomaticComplexity)
     */
    public function getFunnelChartData(
        \DateTime $dateFrom = null,
        \DateTime $dateTo = null,
        Workflow $workflow = null,
        array $customStepCalculations = [],
        AclHelper $aclHelper = null
    ) {
        $data = array();

        if (!$workflow) {
            return $data;
        }

        $steps = $workflow->getStepManager()->getOrderedSteps();

        // regular and final steps should be calculated separately
        $regularSteps = [];
        $finalSteps   = [];
        foreach ($steps as $step) {
            if (!in_array($step->getName(), $this->excludedSteps)) {
                if ($step->isFinal()) {
                    $finalSteps[] = $step->getName();
                } else {
                    $regularSteps[] = $step->getName();
                }
            }
        }

        // regular steps should be calculated for whole period, final steps - for specified period
        $regularStepsData = $this->getStepData($regularSteps, null, null, $customStepCalculations, $aclHelper);
        $finalStepsData   = $this->getStepData($finalSteps, $dateFrom, $dateTo, $customStepCalculations, $aclHelper);

        foreach ($steps as $step) {
            $stepName = $step->getName();
            if (!in_array($stepName, $this->excludedSteps)) {
                $stepLabel = $step->getLabel();
                if ($step->isFinal()) {
                    $dataValue = isset($finalStepsData[$stepName]) ? $finalStepsData[$stepName] : 0;
                    $data[]    = ['value' => $dataValue, 'label' => $stepLabel, 'isNozzle' => true];
                } else {
                    $dataValue = isset($regularStepsData[$stepName]) ? $regularStepsData[$stepName] : 0;
                    $data[]    = ['value' => $dataValue, 'label' => $stepLabel, 'isNozzle' => false];
                }
            }
        }

        return $data;
    }

    /**
     * @param  array     $steps
     * @param  \DateTime $dateFrom
     * @param  \DateTime $dateTo
     * @param  array     $customStepCalculations
     * @param  AclHelper $aclHelper
     *
     * @return array
     */
    protected function getStepData(
        array $steps,
        \DateTime $dateFrom = null,
        \DateTime $dateTo = null,
        array $customStepCalculations = [],
        AclHelper $aclHelper = null
    ) {
        $stepData = [];

        if (!$steps) {
            return $stepData;
        }

        $budgetAmountQueryBuilder = $this->getTemplateQueryBuilder($dateFrom, $dateTo)
            ->addSelect('SUM(opportunity.budgetAmountValue) as budgetAmount');
        $budgetAmountQueryBuilder
            ->andWhere($budgetAmountQueryBuilder->expr()->in('workflowStep.name', ':workflowStepNames'))
            ->setParameter('workflowStepNames', $steps);
        $budgetAmountQuery = $this->getQuery($budgetAmountQueryBuilder, $aclHelper);

        foreach ($budgetAmountQuery->getArrayResult() as $record) {
            $stepData[$record['workflowStepName']] = $record['budgetAmount'] ? (float)$record['budgetAmount'] : 0;
        }

        foreach ($customStepCalculations as $step => $field) {
            if (!in_array($step, $steps)) {
                continue;
            }

            QueryBuilderUtil::checkField($field);
            $customStepQueryBuilder = $this->getTemplateQueryBuilder($dateFrom, $dateTo)
                ->addSelect('SUM(' . $field . ') as value')
                ->andWhere('workflowStep.name = :workflowStep')
                ->setParameter('workflowStep', $step);
            $customStepQueryBuilder
                ->andWhere($customStepQueryBuilder->expr()->in('workflowStep.name', ':workflowStepNames'))
                ->setParameter('workflowStepNames', $steps);
            $customStepQuery = $this->getQuery($customStepQueryBuilder, $aclHelper);

            foreach ($customStepQuery->getArrayResult() as $record) {
                $stepData[$record['workflowStepName']] = $record['value'] ? (float)$record['value'] : 0;
            }
        }

        return $stepData;
    }

    /**
     * @param  QueryBuilder $queryBuilder
     * @param  AclHelper    $aclHelper
     *
     * @return Query
     */
    protected function getQuery(QueryBuilder $queryBuilder, AclHelper $aclHelper = null)
    {
        return $aclHelper ? $aclHelper->apply($queryBuilder) : $queryBuilder->getQuery();
    }

    /**
     * @param  \DateTime $dateFrom
     * @param  \DateTime $dataTo
     *
     * @return QueryBuilder
     */
    protected function getTemplateQueryBuilder(\DateTime $dateFrom = null, \DateTime $dataTo = null)
    {
        $queryBuilder = $this->createQueryBuilder('funnel')
            ->select('workflowStep.name as workflowStepName')
            ->join('funnel.opportunity', 'opportunity');
        $this->joinWorkflowStep($queryBuilder, 'workflowStep');
        $queryBuilder->groupBy('workflowStep.name');

        if ($dateFrom && $dataTo) {
            $queryBuilder
                ->where($queryBuilder->expr()->between('funnel.createdAt', ':dateFrom', ':dateTo'))
                ->setParameter('dateFrom', $dateFrom, Types::DATETIME_MUTABLE)
                ->setParameter('dateTo', $dataTo, Types::DATETIME_MUTABLE);
        }

        return $queryBuilder;
    }
}
